//===============================================================================
// Copyright 2014 Alachisoft.  All rights reserved.
//===============================================================================

using System;
using System.IO;
using System.Diagnostics;
using System.Net;
using System.Text;
using System.Collections;
using System.Reflection;
using System.Runtime.Remoting;
using System.Runtime.InteropServices;
using System.Runtime.Serialization.Formatters.Binary;
using System.Threading;
using System.Management;

using Alachisoft.NCache.Cluster;
using Alachisoft.NCache.Cluster.Util;
using Alachisoft.NCache.Cluster.Stack;

using Alachisoft.NCache.Config;
using Alachisoft.NCache.Caching;
using Alachisoft.NCache.Caching.CacheSchemes;
using Alachisoft.NCache.Caching.CacheSchemes.Local;
using Alachisoft.NCache.Caching.CacheSchemes.Clustered;
using Alachisoft.NCache.Caching.Store;
using Alachisoft.NCache.Caching.Statistics;
using Alachisoft.NCache.Caching.AutoExpiration;
using Alachisoft.NCache.Caching.EvictionPolicies;
using Alachisoft.NCache.Exceptions;


namespace Alachisoft.NCache.Util
{
	/// <summary>
	/// {^[\t, ]*}internal{[a-z, ]*}class{[A-Z,a-z,\,,\:,\t, ]*$}
	/// #if DEBUG\n\1public\2class\3\n#else\n\1internal\2class\3\n#endif
	/// Utility class to help with common tasks.
	/// </summary>
#if DEBUG
	public class MiscUtil
#else
	internal class MiscUtil
#endif
	{ 
		/// <summary>
		/// Writes an error, warning, information, success audit, or failure audit 
		/// entry with the given message text to the event log.
		/// </summary>
		/// <param name="msg">The string to write to the event log.</param>
		/// <param name="type">One of the <c>EventLogEntryType</c> values.</param>
		internal static void LogEvent(string msg, EventLogEntryType type)
		{
			using(EventLog ncLog = new EventLog("Application"))
			{
				ncLog.Source = "NCache";
				ncLog.WriteEntry(msg, type);
			}
		}

		/// <summary>
		/// Returns the number of processors on the system.
		/// </summary>
		public static int NumProcessors
		{
			get
			{
				try
				{
					ManagementClass mc = new ManagementClass("Win32_Processor");
					return mc.GetInstances().Count;
				}
				catch(Exception e)
				{
					Trace.error("MiscUtil.NumProcessors", e.ToString());
				}
				return 0;
			}
		}


		/// <summary>
		/// Returns a list of mac addresses found on the system.
		/// </summary>
		public static ArrayList AdapterAddresses
		{
			get
			{
				ArrayList addrs = new ArrayList();
				try
				{
					ManagementClass mc = new ManagementClass("Win32_NetworkAdapterConfiguration");
					ManagementObjectCollection moc = mc.GetInstances();

					foreach(ManagementObject mo in moc)
					{
						if((bool)mo["IPEnabled"] == true)
						{
							string address = mo["MacAddress"].ToString();
							address = address.Replace(":", "").ToLower();
							addrs.Add(address);
							Trace.info("MiscUtil.AdapterAddresses", "MAC address\t" + address);
						}
						mo.Dispose();
					}
				}
				catch(Exception e)
				{
					Trace.error("MiscUtil.AdapterAddresses", e.ToString());
				}
				return addrs;
			}
		}


		/// <summary>
		/// Returns lg(Log2) of a number.
		/// </summary>
		/// <param name="val"></param>
		/// <returns></returns>
		public static byte Lg(int val)
		{
			byte i = 0;
			while(val > 1)
			{
				val >>= 1;
				i++;
			}
			return i;
		}


		/// <summary>
		/// Converts the IpAddress into a System.Net.IPEndPoint.
		/// </summary>
		/// <param name="address">IpAddress</param>
		/// <returns>System.Net.IPEndPoint</returns>
		public static IPEndPoint AddressToEndPoint(Address address)
		{
			IpAddress ipAddr = address as IpAddress;
			if(ipAddr == null) return null;
			return new IPEndPoint(ipAddr.getIpAddress(), ipAddr.Port);
		}


		/// <summary>
		/// Returns an array containing list of keys contained in the cache. Null if there are no
		/// keys or if timeout occurs.
		/// </summary>
		/// <param name="cache"></param>
		/// <param name="timeout"></param>
		/// <returns></returns>
		public static object[] GetKeyset(CacheBase cache, int timeout)
		{
			ulong index = 0;
			object[] objects = null;
			cache.Sync.AcquireReaderLock(timeout);
			try
			{
				if(!cache.Sync.IsReaderLockHeld || cache.Count < 1) return objects;
				objects = new object[cache.Count];
				for(IEnumerator i = cache.GetEnumerator(); i.MoveNext();)
				{
					objects[index++] = ((DictionaryEntry)i.Current).Key;
				}
			}
			finally
			{
				cache.Sync.ReleaseReaderLock();
			}
			return objects;
		}


		[DllImport("Kernel32.dll")]
		public static extern void QueryPerformanceCounter(ref long ticks);

	}

	/// <summary>
	/// Class that is useful in capturing statistics.
	/// </summary>
#if DEBUG
	[Serializable]
	public class UsageStats
#else
		[Serializable]
		internal class UsageStats
#endif
	{
		/// <summary> Timestamp for the begining of a sample. </summary>
		private long		_lastStart;
		/// <summary> Timestamp for the end of a sample. </summary>
		private long		_lastStop;

		/// <summary>
		/// Constructor
		/// </summary>
		public UsageStats()
		{
			Reset();
		}

		/// <summary>
		/// Returns the time interval for the last sample
		/// </summary>
		public long Current 
		{ 
			get { lock(this){ return _lastStop - _lastStart; } } 
		}

		/// <summary>
		/// Resets the statistics collected so far.
		/// </summary>
		public void Reset()
		{
			_lastStart = _lastStop = 0;
		}

		/// <summary>
		/// Timestamps the start of a sampling interval.
		/// </summary>
		public void BeginSample()
		{
			MiscUtil.QueryPerformanceCounter(ref _lastStart);
		}

			
		/// <summary>
		/// Timestamps the end of interval and calculates the sample time
		/// </summary>
		public void EndSample()
		{
			MiscUtil.QueryPerformanceCounter(ref _lastStop);
		}
	}


	/// <summary>
	/// Class that is useful in capturing statistics.
	/// </summary>
#if DEBUG
	[Serializable]
	public class TimeStats
#else
		[Serializable]
		internal class TimeStats
#endif
	{
		/// <summary> Total number of samples collected for the statistics. </summary>
		private long		_runCount;
		/// <summary> Timestamp for the begining of a sample. </summary>
		private long		_lastStart;
		/// <summary> Timestamp for the end of a sample. </summary>
		private long		_lastStop;
		/// <summary> Total time spent in sampling, i.e., acrued sample time. </summary>
		private long		_totalTime;
		/// <summary> Best time interval mesaured during sampling. </summary>
		private long		_bestTime;
		/// <summary> Worst time interval mesaured during sampling. </summary>
		private long		_worstTime;
		/// <summary> Avg. time interval mesaured during sampling. </summary>
		private float		_avgTime;


		/// <summary>
		/// Constructor
		/// </summary>
		public TimeStats()
		{
			Reset();
		}

		/// <summary>
		/// Returns the total numbre of runs in the statistics capture.
		/// </summary>
		public long Runs	
		{ 
			get { lock(this){ return _runCount; } } 
		}

		/// <summary>
		/// Returns the total time iterval spent in sampling
		/// </summary>
		public long Total	
		{ 
			get { lock(this){ return _totalTime; } } 
		}

		/// <summary>
		/// Returns the time interval for the last sample
		/// </summary>
		public long Current 
		{ 
			get { lock(this){ return _lastStop - _lastStart; } } 
		}

		/// <summary>
		/// Returns the best time interval mesaured during sampling
		/// </summary>
		public long Best	
		{ 
			get { lock(this){ return _bestTime; } } 
			//set { lock(this){ _bestTime = value; } } 
		}

		/// <summary>
		/// Returns the avg. time interval mesaured during sampling
		/// </summary>
		public float Avg		
		{ 
			get { lock(this){ return _avgTime ; } } 
			//set { lock(this){ _avgTime = value; } } 
		}

		/// <summary>
		/// Returns the worst time interval mesaured during sampling
		/// </summary>
		public long Worst	
		{ 
			get { lock(this){ return _worstTime; } } 
			//set { lock(this){ _worstTime = value; } } 
		}

		/// <summary>
		/// Resets the statistics collected so far.
		/// </summary>
		public void Reset()
		{
			_runCount = 0;
			//_lastStart = _lastStop = 0;
			_totalTime = _bestTime = _worstTime = 0;
			_avgTime = 0;
		}

		/// <summary>
		/// Timestamps the start of a sampling interval.
		/// </summary>
		public void BeginSample()
		{
			_lastStart = (DateTime.Now.Ticks - 621355968000000000) / 10000;
		}

			
		/// <summary>
		/// Timestamps the end of interval and calculates the sample time
		/// </summary>
		public void EndSample()
		{
			lock(this)
			{
				_lastStop = (DateTime.Now.Ticks - 621355968000000000) / 10000;
				AddSampleTime(Current);
			}
		}
			
		/// <summary>
		/// Adds a specified sample time to the statistics and updates the run count
		/// </summary>
		/// <param name="time">sample time in milliseconds.</param>
		public void AddSampleTime(long time)
		{
			lock(this)
			{
				_runCount ++;
				if(_runCount == 1)
				{
					_avgTime = _totalTime = _bestTime = _worstTime = time;
				}
				else
				{
					_totalTime += time;
					if(time < _bestTime)	_bestTime = time;
					if(time > _worstTime)	_worstTime = time;
					_avgTime = (float)_totalTime / _runCount;
				}
			}
		}

		/// <summary>
		/// Override converts to string equivalent.
		/// </summary>
		/// <returns></returns>
		public override string ToString()
		{
			lock(this)
			{
				string retval = "[Runs: " + _runCount + ", ";
				retval += "Best(ms): " + _bestTime + ", ";
				retval += "Avg.(ms): " + _avgTime + ", ";
				retval += "Worst(ms): " + _worstTime + "]";
				return retval;
			}
		}
	}


	/// <summary>
	/// class to assist with common system and .NET operations.
	/// </summary>
	internal class Common
	{
		/// <summary>
		/// Creates an instance of the type whose name is specified, using the named 
		/// assembly and the default constructor.
		/// </summary>
		/// <param name="assembly">The name of the assembly where the type named typeName 
		/// is sought. If assemblyName is a null reference, 
		/// the executing assembly is searched. </param>
		/// <param name="classname">The name of the preferred type. </param>
		/// <returns>A reference to the newly created instance.</returns>
		public static object CreateObject(string assembly, string classname)
		{
			return CreateObject(assembly, classname, null);
		}
		
		/// <summary>
		/// Creates an instance of the type whose name is specified, using the named 
		/// assembly and the constructor that best matches the specified parameters.
		/// </summary>
		/// <param name="assembly">The name of the assembly where the type named typeName 
		/// is sought. If assemblyName is a null reference, 
		/// the executing assembly is searched. </param>
		/// <param name="classname">The name of the preferred type. </param>
		/// <param name="args">An array of arguments that match in number, order, and 
		/// type the parameters of the constructor to invoke.</param>
		/// <returns>A reference to the newly created instance.</returns>
		public static object CreateObject(string assembly, string classname, object[] args)
		{
			ObjectHandle objHandle = null;
			try
			{
				objHandle = Activator.CreateInstance(
					assembly,				// name of assembly
					classname,				// fully qualified class name
					false,					// class name is case-sensitive
					BindingFlags.Default,	// no binding attributes	
					null,					// use default binder
					args,					// arguments to constructor,
					null,					// default culture
					null,					// default activation attributes
					null					// default security policy
					);
				if (objHandle != null)
				{
					return objHandle.Unwrap();
				}
			}
			catch(Exception e)
			{
				Trace.error("Common.CreateObject()", e.ToString());
			}
			return null;
		}
																   
		/// <summary>
		/// serialize an objevt into a stream.
		/// </summary>
		/// <param name="stream"></param>
		/// <param name="objectToSend"></param>
		public static bool Serialize(Stream stream, object objectToSend)
		{
			try
			{
				BinaryFormatter formatter = new BinaryFormatter();
				formatter.Serialize(stream, objectToSend);
				return true;
			}
			catch(Exception e)
			{
				Trace.error("Common.Serialize()", e.ToString());
			}
			return false;
		}

		/// <summary>
		/// deserialize an object from the stream.
		/// </summary>
		/// <param name="stream"></param>
		/// <returns></returns>
		public static object Deserialize(Stream stream)
		{
			try
			{
				BinaryFormatter formatter = new BinaryFormatter();
				return formatter.Deserialize(stream);
			}
			catch(Exception e)
			{
				Trace.error("Common.Deserialize()", e.ToString());
			}
			return null;
		}

		/// <summary>
		/// Creates an object from a byte buffer
		/// </summary>
		/// <param name="buffer"></param>
		/// <returns></returns>
		public static object FromByteBuffer(byte[] buffer)
		{
			if(buffer == null) return null;
			return Deserialize(new MemoryStream(buffer));
		}
			
		/// <summary>
		/// Serializes an object into a byte buffer.
		/// The object has to implement interface Serializable or Externalizable
		/// </summary>
		/// <param name="obj"></param>
		/// <returns></returns>
		public static byte[] ToByteBuffer(object obj)
		{
			if(obj == null) return null;
			//if(obj.GetType().Equals(typeof(byte[]))) return obj as byte[];
			MemoryStream ms = new MemoryStream();
			try
			{
				Serialize(ms, obj);
				byte[] buf = new byte[ms.Length];
				ms.Position = 0;
				ms.Read(buf,0,buf.Length);
				ms.Close();
				return buf;
			}
			catch(Exception e)
			{
				Trace.error("Common.ToByteBuffer()", e.ToString());
			}
			return null;
		}		
	}

	internal class Trace
	{
		private static bool isInfoEnabled = true;
		private static bool isErrorEnabled = true;
		private static bool isWarnEnabled = true;
		private static bool isDebugEnabled = true;
		private static bool isFatalEnabled = true;

		/// <summary>
		/// Information Trace
		/// </summary>
		/// <param name="module">Module responsible for the information</param>
		/// <param name="message">The message to be displayed</param>
		[Conditional("DEBUG")] 
		public static void info(String module, String message) 
		{
			if(isInfoEnabled)
				writeToDebug("[info]",module,message);
		}

		[Conditional("DEBUG")] 
		public static void info(String message) 
		{
			info("", message);
		}

		/// <summary>
		/// Warning Trace
		/// </summary>
		/// <param name="module">Module responsible for the warning</param>
		/// <param name="message">The message to be displayed</param>
		[Conditional("DEBUG")] 
		public static void warn(String module, String message) 
		{
			if(isWarnEnabled)
				writeToDebug("[WARN]",module,message);
		} 
	
		[Conditional("DEBUG")] 
		public static void warn(String message) 
		{
			warn("",message);
		}

		/// <summary>
		/// Error Trace
		/// </summary>
		/// <param name="module">Module responsible for the error</param>
		/// <param name="message">The message to be displayed</param>
		[Conditional("DEBUG")] 
		public static void error(String module, String message) 
		{
			if(isErrorEnabled)
				writeToDebug("[ERR]",module,message);
		}

		[Conditional("DEBUG")] 
		public static void error(String message) 
		{
			error("",message);
		}

		[Conditional("DEBUG")] 
		public static void fatal(String module, String message) 
		{
			if(isFatalEnabled)
				writeToDebug("[FATAL]",module,message);
		}

		[Conditional("DEBUG")] 
		public static void fatal(String message) 
		{
			fatal("",message);
		}

		[Conditional("DEBUG")] 
		public static void debug(String module, String message) 
		{
			if(isDebugEnabled)
				writeToDebug("[DBG]","",message);
		}

		[Conditional("DEBUG")] 
		public static void debug(String message) 
		{
			debug("",message);
		}

		/// <summary>
		/// Writes the trace to the Debug
		/// </summary>
		/// <param name="type">Type of trace</param>
		/// <param name="module">Module responsible for the error</param>
		/// <param name="message">The message to be displayed</param>
		[Conditional("DEBUG")] 
		private static void writeToDebug(String type, String module, String message)
		{
			int space1 = 8;
			int space2 = 40;
			Debug.WriteLine(System.DateTime.Now.ToString("HH:mm:ss:ff") + "      " +  type.PadRight(space1,' ') + module.PadRight(space2,' ') + message);
			//System.Diagnostics.Trace.WriteLine(System.DateTime.Now.ToString("HH:mm:ss:ff") + "      " +  type.PadRight(space1,' ') + module.PadRight(space2,' ') + message);
			//System.Console.WriteLine(type + " " + message);
		}
	}
}
